/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* ROUTINES TO READ THE GROUP FILE
 *
 * Can be Unix group database or our own group file.
 *
 *  Our group database file format is similar to the Unix one:
 *     group_name::GID:user_list
 *  Where user_list is a comma separated list of users who are in that group,
 *  (not necessarily including those who have it as their primary group id).
 *  Long lines may be broken up with \ at end of each broken line.
 */

#include "backtalk.h"

#ifdef REAL_GID
#include <grp.h>
#endif
#include <ctype.h>

#include "group.h"

#ifdef REAL_GID

/* GROUPNAME - Given a gid number, return the matching group name.  Return
 * NULL if not defined.  Returned name should be treated as being in static
 * memory - future calls may overwrite it.
 *
 * Unix version.
 */

char *groupname(gid_t gid)
{
struct group *gr;

    setgrent();
    while ((gr= getgrent()) != NULL)
    {
    	if (gr->gr_gid == gid)
	{
	    endgrent();
	    return gr->gr_name;
	}
    }
    endgrent();
    return NULL;
}

/* GROUPID - Given a group name, return the matching group id.  Return
 * -1 if not defined.
 *
 * Unix version.
 */

int groupid(char *gname)
{
struct group *gr;

    setgrent();
    while ((gr= getgrent()) != NULL)
    {
    	if (!strcmp(gr->gr_name,gname))
	{
	    endgrent();
	    return gr->gr_gid;
	}
    }
    endgrent();
    return -1;
}


/* INGROUPNO - Given a group number return true if the given user with the
 * given primary group id is in that group.
 *
 * Unix version.
 */

int ingroupno(gid_t gid, gid_t my_gid, char *my_login)
{
struct group *gr;
char **m;

    if (my_login == NULL || my_login[0] == '\0') return 0;
    if (gid == my_gid) return 1;

    setgrent();
    while ((gr= getgrent()) != NULL)
    {
    	if (gr->gr_gid == gid)
    	{
	    for (m= gr->gr_mem; *m != NULL; m++)
	    {
	    	if (!strcmp(*m, my_login))
		{
		    endgrent();
		    return 2;
	        }
	    }
	    endgrent();
	    return 0;
	}
    }
    endgrent();
    return 0;
}


/* INGROUPNAME - Given a group name return true if the given user with the 
 * given group id is in that group.
 *
 * Unix version.
 */

int ingroupname(char *groupname, gid_t my_gid, char *my_login)
{
struct group *gr;
char **m;

    if (groupname == NULL || groupname[0] == '\0') return 0;
    if (my_login == NULL || my_login[0] == '\0') return 0;

    setgrent();
    while ((gr= getgrent()) != NULL)
    {
    	if (!strcmp(groupname, gr->gr_name))
    	{
	    if (gr->gr_gid == my_gid) return 1;
	    for (m= gr->gr_mem; *m != NULL; m++)
	    {
	    	if (!strcmp(*m, my_login))
		{
		    endgrent();
		    return 2;
	        }
	    }
	    endgrent();
	    return 0;
	}
    }
    endgrent();
    return 0;
}


/* INGROUPLIST - Given a NULL-terminated array of pointers to group names and
 * an ID, return true if that id is in any of those groups.
 *
 * Unix version.
 */

int ingrouplist(char **groups, gid_t my_gid, char *my_login)
{
struct group *gr;
char **m;
int i;

    if (groups == NULL || groups[0] == NULL) return 0;

    setgrent();
    while ((gr= getgrent()) != NULL)
    {
	for (i= 0; groups[i] != NULL; i++)
	{
	    if (!strcmp(groups[i], gr->gr_name))
	    {
		if (gr->gr_gid == my_gid) return 1;
		for (m= gr->gr_mem; *m != NULL; m++)
		{
		    if (!strcmp(*m, my_login))
		    {
			endgrent();
			return 1;
		    }
		}
		endgrent();
	    }
	}
    }
    endgrent();
    return 0;
}


/* FOR_GROUPS - Call a given function once with each currently defined group
 * name.
 *
 * Unix version.
 */

void for_groups(void (*func)(char *))
{
struct group *gr;

    setgrent();
    while ((gr= getgrent()) != NULL)
    	(*func)(gr->gr_name);
    endgrent();
}

#else

/* GROUPNAME - Given a gid number, return the matching group name.  Return
 * NULL if not defined.  Returned name should be treated as being in static
 * memory - future calls may overwrite it.
 *
 * File version.
 */

char *groupname(gid_t gid)
{
FILE *fp;
static char bf[BFSZ+1];
char *p, *q;
int continued= 0;
int len;

    if ((fp= fopen(GROUP_FILE,"r")) == NULL)
    {
	/* No file - use defaults */
#if defined(CFADM_GID) && defined(CFADM_GROUP)
	if (gid == CFADM_GID) return CFADM_GROUP;
#endif
#if defined(USER_GID) && defined(USER_GROUP)
	if (gid == USER_GID) return USER_GROUP;
#endif
#if defined(GRADM_GID) && defined(GRADM_GROUP)
	if (gid == GRADM_GID) return GRADM_GROUP;
#endif
    	return NULL;
    }

    while (fgets(bf, BFSZ, fp) != NULL)
    {
	len= strlen(bf);
    	if (!continued && bf[0] != '#' && (p= strchr(bf,':')) != NULL &&
	    (q= strchr(p+1,':')) != NULL && isascii(q[1]) && isdigit(q[1]) &&
	    atoi(q+1) == gid)
	{
	    *p= '\0';
	    fclose(fp);
	    return bf;
	}
	continued= ((bf[len-1] != '\n') || (bf[len-2] == '\\'));
    }
    fclose(fp);
    return NULL;
}


/* GROUPID - Given a group name, return the matching group id.  Return
 * -1 if not defined.
 *
 * File version.
 */

int groupid(char *gname)
{
FILE *fp;
static char bf[BFSZ+1];
char *p, *q;
int continued= 0;
int len;

    if ((fp= fopen(GROUP_FILE,"r")) == NULL)
    {
	/* No file - use defaults */
#if defined(CFADM_GID) && defined(CFADM_GROUP)
	if (!strcmp(gname,CFADM_GROUP)) return CFADM_GID;
#endif
#if defined(USER_GID) && defined(USER_GROUP)
	if (!strcmp(gname,USER_GROUP)) return USER_GID;
#endif
#if defined(GRADM_GID) && defined(GRADM_GROUP)
	if (!strcmp(gname,GRADM_GROUP)) return GRADM_GID;
#endif
    	return -1;
    }

    while (fgets(bf, BFSZ, fp) != NULL)
    {
	len= strlen(bf);
    	if (!continued && bf[0] != '#' && (p= strchr(bf,':')) != NULL)
	{
	    *p= '\0';
	    if (!strcmp(bf,gname))
	    {
		fclose(fp);
		if ((q= strchr(p+1,':')) == NULL) return -1;
		return atoi(q+1);
	    }
	}
	continued= ((bf[len-1] != '\n') || (bf[len-2] == '\\'));
    }
    fclose(fp);
    return -1;
}


/* INGROUPNO - Given a group number return true if the given user with the
 * given group id is in that group.
 *
 * File version - do character-at-a-time reading to handle really long lines
 * without buffer overflow.
 */

int ingroupno(gid_t gid, gid_t my_gid, char *my_login)
{
FILE *fp;
char bf[BFSZ+1];
int state, ch, i, esc;

    if (my_login == NULL || my_login[0] == '\0') return 0;
    if (gid < 0) return 0;
    if (gid == my_gid) return 1;

    if ((fp= fopen(GROUP_FILE,"r")) == NULL)
	return 0;

    state= 0;
    esc= 0;
    while ((ch= fgetc(fp)) != EOF)
    {
    	switch (state)
	{
	case -1: /* skip to start of line */
	    if (ch == '\n' && !esc) state= 0;
	    break;

	case 0: /* ready to read group name */
	    if (ch == '#') {state= -1; break;}
	    state= 1;
	    i= 0;	/* count of colons */
	    /* drop through */

	case 1: /* reading group name */
	    if (ch == ':')
	    {
		/* on second colon - start reading gid */
		if (++i == 2) state= 2;
	    }
	    else if (ch == '\n' && !esc)
		state= 0;
	    break;

	case 2: /* ready to read gid */
	    state= 3;
	    i= 0;
	    /* drop through */

	case 3: /* reading gid number */
	    if (ch == ':' || (ch == '\n' && !esc))
	    {
		bf[i]= '\0';
		if (i <= 0 || !isascii(bf[0]) || !isdigit(bf[0]) ||
		         atoi(bf) != gid)
		    state= -1;	/* gid doesn't match - skip to next line */
		else if (ch != '\n')
		    state= 4;	/* gid matches - check names */
		else
		{
		    /* found line, but gid doesn't match and no names */
		    fclose(fp);
		    return 0;
		}
	    }
	    else if (ch == '\n')	/* line ends with \ */
	    {
		if (i > 0) i--;
	    }
	    else
	    {
	    	bf[i++]= ch;
	    	if (i >= BFSZ) state= -1;
	    }
	    break;

	case 4: /* ready to read a name */
	    state= 5;
	    i= 0;

	case 5: /* reading a name */
	    if (ch == ',' || (ch == '\n' && !esc))
	    {
	    	bf[i]= '\0';
		if (!strcmp(my_login, bf))
		{
		    fclose(fp);
		    return 2;
		}

		if (ch == '\n')
		{
		    /* no more names */
		    fclose(fp);
		    return 0;
		}
		else
		    state= 4;	/* read next name */
	    }
	    else if (ch == '\n')	/* line ends with \ */
	    {
		if (i > 0) i--;
	    }
	    else
	    {
	    	bf[i++]= ch;
	    	if (i >= BFSZ)
		{
		    fclose(fp);
		    return 0;
		}
	    }
	    break;
	}
	esc= (ch == '\\');
    }

    /* group not found */
    fclose(fp);
    return 0;
}


/* INGROUPNAME - Given a group name return true if the given user with the
 * given group id is in that group.
 *
 * File version - do character-at-a-time reading to handle really long lines
 * without buffer overflow.
 */

int ingroupname(char *groupname, gid_t my_gid, char *my_login)
{
FILE *fp;
char bf[BFSZ+1];
int state, ch, i, esc;

    if (groupname == NULL || groupname[0] == '\0') return 0;
    if (my_login == NULL || my_login[0] == '\0') return 0;

    if ((fp= fopen(GROUP_FILE,"r")) == NULL)
    {
	/* No file - use defaults */
	return
#if defined(CFADM_GID) && defined(CFADM_GROUP)
	    (my_gid == CFADM_GID && !strcmp(groupname,CFADM_GROUP)) ||
#endif
#if defined(USER_GID) && defined(USER_GROUP)
	    (my_gid == USER_GID && !strcmp(groupname,USER_GROUP)) ||
#endif
#if defined(GRADM_GID) && defined(GRADM_GROUP)
	    (my_gid == GRADM_GID && !strcmp(groupname,GRADM_GROUP)) ||
#endif
	    0;
    }

    state= 0;
    esc= 0;
    while ((ch= fgetc(fp)) != EOF)
    {
    	switch (state)
	{
	case -1: /* skip to start of line */
	    if (ch == '\n' && !esc) state= 0;
	    break;

	case 0: /* ready to read group name */
	    if (ch == '#') {state= -1; break;}
	    state= 1;
	    i= 0;
	    /* drop through */

	case 1: /* reading group name */
	    if (ch == ':')
	    {
		bf[i]= '\0';
		if (strcmp(bf, groupname))
		    state= -1;	/* group name doesn't match - goto next line */
		else
		    state= 10;	/* group name matches - read gid */
	    }
	    else if (ch == '\n')
	    {
		if (!esc)
		    state= 0;
		else if (i > 0)
		    i--;
	    }
	    else
	    {
		bf[i++]= ch;
		if (i >= BFSZ) state= -1;
	    }
	    break;

	case 10: /* skip to next colon (OK, the state number is out of seq) */
	    if (ch == ':') state= 2;
	    break;

	case 2: /* ready to read gid */
	    state= 3;
	    i= 0;
	    /* drop through */

	case 3: /* reading gid number */
	    if (ch == ':' || (ch == '\n' && !esc))
	    {
		bf[i]= '\0';
		/* check gid */
		if (my_gid == atoi(bf))
		{
		    fclose(fp);
		    return 1;
		}
		state= 4;	/* check names */
	    }
	    else if (ch == '\n')	/* line ends with \ */
	    {
		if (i > 0) i--;
	    }
	    else
	    {
	    	bf[i++]= ch;
	    	if (i >= BFSZ) state= -1;
	    }
	    break;

	case 4: /* ready to read a name */
	    state= 5;
	    i= 0;

	case 5: /* reading a name */
	    if (ch == ',' || (ch == '\n' && !esc))
	    {
	    	bf[i]= '\0';
		if (!strcmp(my_login, bf))
		{
		    fclose(fp);
		    return 2;
		}

		if (ch == '\n')
		{
		    /* no more names */
		    fclose(fp);
		    return 0;
		}
		else
		    state= 4;	/* read next name */
	    }
	    else if (ch == '\n')	/* line ends with \ */
	    {
		if (i > 0) i--;
	    }
	    else
	    {
	    	bf[i++]= ch;
	    	if (i >= BFSZ)
		{
		    fclose(fp);
		    return 0;
		}
	    }
	    break;
	}
	esc= (ch == '\\');
    }

    /* group not found */
    fclose(fp);
    return 0;
}

/* INGROUPLIST - Given a NULL-terminated array of pointers to group names and
 * an ID, return true if that id is in any of those groups.
 *
 * File version.
 */

int ingrouplist(char **groups, gid_t my_gid, char *my_login)
{
FILE *fp;
char bf[BFSZ+1];
int state, ch, i, esc, j;

    if (groups == NULL || groups[0] == NULL) return 0;

    if ((fp= fopen(GROUP_FILE,"r")) == NULL)
    {
	/* No file - use defaults */
	for (i= 0; groups[0] != NULL; i++)
	{
#if defined(CFADM_GID) && defined(CFADM_GROUP)
	    if (my_gid == CFADM_GID && !strcmp(groups[i],CFADM_GROUP))
	    	return 1;
#endif
#if defined(USER_GID) && defined(USER_GROUP)
	    if (my_gid == USER_GID && !strcmp(groups[i],USER_GROUP))
	    	return 1;
#endif
#if defined(GRADM_GID) && defined(GRADM_GROUP)
	    if (my_gid == GRADM_GID && !strcmp(groups[i],GRADM_GROUP))
	    	return 1;
#endif
	}
	return 0;
    }

    state= 0;
    esc= 0;
    while ((ch= fgetc(fp)) != EOF)
    {
    	switch (state)
	{
	case -1: /* skip to start of line */
	    if (ch == '\n' && !esc) state= 0;
	    break;

	case 0: /* ready to read group name */
	    if (ch == '#') {state= -1; break;}
	    state= 1;
	    i= 0;
	    /* drop through */

	case 1: /* reading group name */
	    if (ch == ':')
	    {
		bf[i]= '\0';
		state= -1;   /* if group name doesn't match - goto next line */
		for (j= 0; groups[j] != NULL; j++)
		{
		    if (!strcmp(bf, groups[j]))
		    {
			state= 10;	/* group name matches - read gid */
			break;
		    }
		}
	    }
	    else if (ch == '\n')
	    {
		if (!esc)
		    state= 0;
		else if (i > 0)
		    i--;
	    }
	    else
	    {
		bf[i++]= ch;
		if (i >= BFSZ) state= -1;
	    }
	    break;

	case 10: /* skip to next colon (OK, the state number is out of seq) */
	    if (ch == ':') state= 2;
	    break;

	case 2: /* ready to read gid */
	    state= 3;
	    i= 0;
	    /* drop through */

	case 3: /* reading gid number */
	    if (ch == ':' || (ch == '\n' && !esc))
	    {
		bf[i]= '\0';
		/* check gid */
		if (my_gid == atoi(bf))
		{
		    fclose(fp);
		    return 1;
		}
		state= 4;	/* check names */
	    }
	    else if (ch == '\n')	/* line ends with \ */
	    {
		if (i > 0) i--;
	    }
	    else
	    {
	    	bf[i++]= ch;
	    	if (i >= BFSZ) state= -1;
	    }
	    break;

	case 4: /* ready to read a name */
	    state= 5;
	    i= 0;

	case 5: /* reading a name */
	    if (ch == ',' || (ch == '\n' && !esc))
	    {
	    	bf[i]= '\0';
		if (!strcmp(my_login, bf))
		{
		    fclose(fp);
		    return 1;
		}

		if (ch == '\n')
		{
		    state= 0;
		}
		else
		    state= 4;	/* read next name */
	    }
	    else if (ch == '\n')	/* line ends with \ */
	    {
		if (i > 0) i--;
	    }
	    else
	    {
	    	bf[i++]= ch;
	    	if (i >= BFSZ)
		{
		    fclose(fp);
		    return 0;
		}
	    }
	    break;
	}
	esc= (ch == '\\');
    }

    /* group not found */
    fclose(fp);
    return 0;
}

/* FOR_GROUPS - Call a given function once with each currently defined group
 * name.
 *
 * File version.
 */

void for_groups(void (*func)(char *))
{
FILE *fp;
static char bf[BFSZ+1];
char *p, *q;
int continued= 0;
int len;

    if ((fp= fopen(GROUP_FILE,"r")) == NULL)
    {
	/* No file - use defaults */
#ifdef CFADM_GROUP
	(*func)(CFADM_GROUP);
#endif
#if defined(USER_GID) && defined(USER_GROUP)
	(*func)(USER_GROUP);
#endif
#if defined(GRADM_GID) && defined(GRADM_GROUP)
	(*func)(GRADM_GROUP);
#endif
    	return;
    }

    while (fgets(bf, BFSZ, fp) != NULL)
    {
	len= strlen(bf);
    	if (!continued && bf[0] != '#' && (p= strchr(bf,':')) != NULL)
	{
	    *p= '\0';
	    (*func)(bf);
	}
	continued= ((bf[len-1] != '\n') || (bf[len-2] == '\\'));
    }
    fclose(fp);
}

#endif /* REAL_GID */


/* USER_GROUP_NAME()
 * USER_GROUP_ID()
 * CFADM_GROUP_NAME()
 * CFADM_GROUP_ID()
 * GRADM_GROUP_NAME()
 * GRADM_GROUP_ID()
 *   Return names/IDs of interesting groups.  Sometimes these are macros
 *   defined in group.h.  Look up each thing only once, then save in a
 *   static variable.
 */

#if defined(USER_GID) && !defined(USER_GROUP)
char *user_group_name()
{
static char *name= NULL;
    if (name == NULL) name= strdup(groupname(USER_GID));
    return name;
}
#endif

#if !defined(USER_GID) && defined(USER_GROUP)
int user_group_id()
{
static int gid= -1;
    if (gid == -1) gid= groupid(USER_GROUP);
    return gid;
}
#endif

#if defined(CFADM_GID) && !defined(CFADM_GROUP)
char *cfadm_group_name()
{
static char *name= NULL;
    if (name == NULL) name= strdup(groupname(CFADM_GID));
    return name;
}
#endif

#if !defined(CFADM_GID) && defined(CFADM_GROUP)
int cfadm_group_id()
{
static int gid= -1;
    if (gid == -1) gid= groupid(CFADM_GROUP);
    return gid;
}
#endif

#if defined(GRADM_GID) && !defined(GRADM_GROUP)
char *gradm_group_name()
{
static char *name= NULL;
    if (name == NULL) name= strdup(groupname(GRADM_GID));
    return name;
}
#endif

#if !defined(GRADM_GID) && defined(GRADM_GROUP)
int gradm_group_id()
{
static int gid= -1;
    if (gid == -1) gid= groupid(GRADM_GROUP);
    return gid;
}
#endif
