/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* USER INFORMATION
 *
 * This file contains routines to access information about the user.  Many
 * sources are possible.  The system may be configured to use either or both
 * of two methods to identify the user:
 *
 *     ID_GETUID
 *        Primary user identification is taken from the getuid() call.  This
 *	  uid is used to directly access the other databases, unless
 *	  UID_MAP_FILE is defined, in which case it is mapped to a user name
 *	  using this file.
 *     ID_REMOTE_USER
 *        Primary user identification is taken from the "REMOTE_USER" 
 *	  environment variable, if and only iff getuid() returns SERVER_UID.
 *
 * The ID_GETUID define can be used with lynxcgi.  ID_REMOTE_USER is used
 * with standard .htaccess-like server authentication schemes.  If both are
 * defined, getuid() is used as the identification unless it returns
 * SERVER_UID, in which case the "REMOTE_USER" environment variable is used.
 * If neither is defined, authentication fails for all users, so they all
 * appear anonymous.
 *
 * Once we have the primary identification, this is used to look up all other
 * user information.  This maybe done using several configurations.  Only one
 * of the following defines may be enabled:
 *
 *  UD_PASSWD
 *      Secondary User Data is taken from the /etc/passwd file.  If
 *	PASSWD_FILE is defined, it is a file of the same format as the
 *	/etc/passwd_file to be used instead.
 *
 *  UD_DBM
 *      Secondary User Data is taken from the DBM file DBM_NAM_FILE or
 *      DBM_UID_FILE, depending on whether we have a name or a uid as our
 *	primary key.  The dbm file is keyed on the primary id.  The content
 *	looks like:
 *   
 *       <passwd>:<otherid>:<gid>:<fullname>:<directory>:<ignored>
 *
 *	Where
 *		<passwd> = encrypted user password.
 *		<otherid> = uid or id -- whichever was not the key.
 *		<gid> = group-id
 *		<fullname> = user's full name.
 *		<directory> = user's home directory.
 *		<ignored> = skip this field for now.
 *
 *	This could be the same file used as Apache's AuthDBMUserFile
 *	Actually, it would be weird to use this with ID_GETUID.
 *
 *  UD_TAGFILE
 *      There is no secondary user information database.  This does not work
 *	unless the primary ID is a name.  The user's directory is constructed
 *	directly from the user name using USER_DIR and USER_DIR_LEVEL.  All
 *	other information is taken from the IDENT_FILE maintained there by
 *	backtalk.  This configuration should not be used with real unix
 *	home directories.
 */


#include "common.h"

#include <pwd.h>

#include "user.h"
#include "udb_read.h"
#include "group.h"

struct passwd *upw;		/* NULL means user is anonymous */
char *new_fname= NULL;		/* If fullname has changed, this is it */


/* DUP_PASSWD - Given a pointer to a 'struct passwd' duplicate it into
 * dynamically allocated memory and return a pointer to the duplicate.
 */

struct passwd *dup_passwd(struct passwd *old)
{
struct passwd *new;
    if (old == NULL) return NULL;

    new= (struct passwd *)malloc(sizeof(struct passwd));

    new->pw_name= old->pw_name ? strdup(old->pw_name) : NULL;
    new->pw_passwd= old->pw_passwd ? strdup(old->pw_passwd) : NULL;
    new->pw_uid= old->pw_uid;
    new->pw_gid= old->pw_gid;
    new->pw_gecos= old->pw_gecos ? strdup(old->pw_gecos) : NULL;
    new->pw_dir= old->pw_dir ? strdup(old->pw_dir) : NULL;
#if 0
    new->pw_shell= old->pw_shell ? strdup(old->pw_shell) : NULL;
#else
    new->pw_shell= NULL;
#endif
    return new;
}


/* GETUSER - Identify the user and save his database informtion into the
 * upw variable.  On failure, we set upw to NULL.  We return a error code:
 *   0 - success
 *   1 - failure - no authentication - should treat as anonymous user
 *   2 - failure - authentication expired - should re-authenticate
 * This must be called before any call to user_fname(), user_id(), user_dir(),
 * user_uid() or user_gid() and before calls to func_changename(),
 * func_changepass() or func_removeuser() which apply to the current user.
 */

int getuser()
{
char *nam;
uid_t realuid;
int rc= 1;
#ifdef ID_SESSION
char bf[MAX_LOGIN_LEN+1];
#endif

    realuid= getuid();
    upw= NULL;		/* Yes, this is needed! */

#if defined(ID_REMOTE_USER) || defined(ID_SESSION)
#ifndef INSECURE
    if (realuid == SERVER_UID || realuid == 0)
#endif
    {
#ifdef ID_REMOTE_USER
	nam= getenv("REMOTE_USER");
	if (nam == NULL)
#endif
	{
#ifdef ID_SESSION
	    if ((rc= session_auth(nam= bf)) != 0)
#endif
		return rc;
	}

	upw= dup_passwd(getdbnam(nam));
	return (upw == NULL);
    }
#endif /* ID_REMOTE_USER || ID_SESSION */

#ifdef ID_GETUID
#ifdef UID_MAP_FILE
    if ((nam= mapuid(realuid)) != NULL)
	upw= dup_passwd(getdbnam(nam));
#else
    upw= dup_passwd(getdbuid(realuid));
#endif /* UID_MAP_FILE */
#endif /* ID_GETUID */
    return (upw == NULL);
}



/* USER_FNAME - Return full name of the user.  Returns "" if hte user is
 * anonymous.
 */

char *user_fname()
{
    return (new_fname == NULL ?
		(upw == NULL ?  "" : upw->pw_gecos) : new_fname);
}


/* USER_ID - Return login id of the user.  Returns "" if the user is anonymous.
 */

char *user_id()
{
    return (upw == NULL ? "" : upw->pw_name);
}


/* USER_DIR - Return home directory of the user.  Returns "" if the user is
 * anonymous.
 */

char *user_dir()
{
    return (upw == NULL ? "" : upw->pw_dir);
}


/* USER_UID - Return uid of the user.  Returns -1 if he is anonymous.
 */

int user_uid()
{
    return (upw == NULL ? -1 : upw->pw_uid);
}


/* USER_GID - Return primary group id of the user.  Returns -1 if he is
 * anonymous.
 */

int user_gid()
{
    return (upw == NULL ? -1 : upw->pw_gid);
}

/* USER_GNAME - Return the primary group name for the user.  Returns NULL if
 * he is anonymous.
 */

char *user_gname()
{
static char *gname= NULL;
int gid= user_gid();
    if (gname != NULL) return gname;
    if (gid == -1) return NULL;
    return gname= groupname(gid);
}


/* USER_CFADM - Return true if the user is in the CFADM group  Return value
 * if 1 if it is our primary group number, 2 if it is a secondary group,
 * and 0 if it is neither.
 */

int user_cfadm()
{
    if (upw == NULL) return 0;
#ifdef CFADM_GID
    return ingroupno(CFADM_GID,upw->pw_gid,upw->pw_name);
#else
#ifdef CFADM_GROUP
    return ingroupname(CFADM_GROUP,upw->pw_gid,upw->pw_name);
#else
    return 0;
#endif
#endif
}


/* USER_GRADM - Return true if the user is in the GRADM group.  Return
 * value is 1 if it is our primary group number, 2 if it is a secondary group,
 * and 0 if it is neither.
 */

int user_gradm()
{
    if (upw == NULL) return 0;
#ifdef GRADM_GID
    return ingroupno(GRADM_GID,upw->pw_gid,upw->pw_name);
#else
#ifdef GRADM_GROUP
    return ingroupname(GRADM_GROUP,upw->pw_gid,upw->pw_name);
#else
    return 0;
#endif
#endif
}
